package net.h34t.filemure.repository

import net.h34t.filemure.ExtId
import net.h34t.filemure.core.entity.*
import net.h34t.filemure.core.entity.Tag.Companion.serialize
import java.io.InputStream
import java.sql.Connection
import java.sql.DriverManager
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter

class SqliteRepository(url: String) {

    private val dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")

    val connection: Connection = DriverManager.getConnection(url)

    fun addFileToLimbo(accountId: Long, filename: String, contentType: String?, size: Long, content: InputStream) {
        connection.prepareStatement("INSERT INTO file (account_id, ext_id, filename, content_type, file_size, content) VALUES (?,?,?,?,?,?)")
            .use { stmt ->
                stmt.setLong(1, accountId)
                stmt.setString(2, ExtId.generate().toString())
                stmt.setString(3, filename)
                stmt.setString(4, contentType)
                stmt.setLong(5, size)
                stmt.setBytes(6, content.readAllBytes())

                val res = stmt.executeUpdate()

                require(res == 1)
            }
    }

    fun getLimboFileCount(accountId: Long, state: State = State.ACTIVE): Long {
        connection.prepareStatement("SELECT count(*) AS count FROM file WHERE account_id=? AND document_id IS NULL AND state=?")
            .use { stmt ->
                stmt.setLong(1, accountId)
                stmt.setInt(2, state.code)
                val rs = stmt.executeQuery()
                rs.next()
                return rs.getLong(1)
            }
    }

    fun getFilesInLimbo(accountId: Long, state: State = State.ACTIVE): List<FileRef> {
        connection.prepareStatement(
            """
            |SELECT 
            |   id, 
            |   ext_id, 
            |   account_id, 
            |   filename, 
            |   content_type, 
            |   content_extracted, 
            |   file_size, 
            |   created,
            |   state
            |FROM 
            |   file 
            |WHERE 
            |   account_id=? AND 
            |   document_id IS NULL 
            |   AND state=? 
            |ORDER BY 
            |   created DESC
            """.trimMargin()
        ).use { stmt ->
            stmt.setLong(1, accountId)
            stmt.setInt(2, state.code)
            val res = stmt.executeQuery()

            val list = mutableListOf<FileRef>()

            while (res.next()) {
                list += FileRef(
                    id = res.getLong("id"),
                    extId = res.getString("ext_id"),
                    accountId = res.getLong("account_id"),
                    documentId = null,
                    filename = res.getString("filename"),
                    contentType = res.getString("content_type"),
                    contentExtracted = res.getString("content_extracted"),
                    fileSize = res.getLong("file_size"),
                    created = LocalDateTime.parse(res.getString("created"), dtf),
                    state = State.fromCode(res.getInt("state"))
                )
            }

            return list
        }
    }

    fun addDocument(
        accountId: Long,
        title: String,
        referenceDate: LocalDateTime,
        tags: List<String>,
        description: String,
        fileExtIds: List<String>
    ): ExtId {

        val savePoint = connection.setSavepoint()
        try {
            val extId = ExtId.generate()

            val documentId = connection.prepareStatement(
                """INSERT INTO document 
            |(account_id, ext_id, title, description, tags, created, reference_date) 
            |VALUES 
            |(?, ?, ?, ?, ?, datetime(), ?)""".trimMargin()
            ).use { stmt ->
                stmt.setLong(1, accountId)
                stmt.setString(2, extId.value)
                stmt.setString(3, title)
                stmt.setString(4, description)
                stmt.setString(5, tags.joinToString(","))
                stmt.setString(6, referenceDate.format(dtf))
                stmt.executeUpdate()
                val gks = stmt.generatedKeys
                gks.next()
                gks.getLong(1)
            }

            val extIds = fileExtIds.joinToString(",") { "'$it'" }

            connection.prepareStatement("""UPDATE file SET document_id=? WHERE account_id=? AND ext_id IN ($extIds)""")
                .use { stmt ->
                    stmt.setLong(1, documentId)
                    stmt.setLong(2, accountId)
                    val affected = stmt.executeUpdate()
                    require(affected == fileExtIds.size)
                }

            connection.commit()
            return extId
        } catch (exception: Exception) {
            connection.rollback(savePoint)
            throw exception
        }
    }

    fun getDocuments(accountId: Long, state: State = State.ACTIVE): List<Document> {
        connection.prepareStatement(
            """
            |SELECT 
            |   d.id, 
            |   d.account_id, 
            |   d.ext_id, 
            |   d.title, 
            |   d.description, 
            |   d.tags, 
            |   d.created, 
            |   d.reference_date,
            |   d.state
            |FROM 
            |   document d 
            |WHERE 
            |   account_id=? AND
            |   state=?
            """.trimMargin()
        ).use { stmt ->
            stmt.setLong(1, accountId)
            stmt.setInt(2, state.code)
            val res = stmt.executeQuery()

            val documentList = mutableListOf<Document>()

            while (res.next()) {
                documentList.add(
                    Document(
                        id = res.getLong("id"),
                        extId = res.getString("ext_id"),
                        title = res.getString("title"),
                        description = res.getString("description"),
                        tags = res.getString("tags")
                            ?.let { if (it.isNotBlank()) it.split(",").map { tag -> Tag(tag) } else emptyList() }
                            ?: emptyList(),
                        created = LocalDateTime.parse(res.getString("created"), dtf),
                        referenceDate = LocalDateTime.parse(res.getString("reference_date"), dtf),
                        state = State.fromCode(res.getInt("state")),
                        files = emptyList()
                    )
                )
            }


            return documentList.toList()
        }
    }

    fun updateDocument(
        accountId: Long,
        id: Long,
        title: String,
        referenceDate: LocalDateTime,
        tags: List<Tag>,
        description: String,
        state: State = State.ACTIVE,
    ) {
        connection.prepareStatement(
            """
            |UPDATE
            |   document
            |SET
            |   title=?,
            |   reference_date=?,
            |   tags=?,
            |   description=?,
            |   state=?
            |WHERE
            |   account_id=? AND
            |   id=?
        """.trimMargin()
        ).use { stmt ->
            stmt.setString(1, title)
            stmt.setString(2, dtf.format(referenceDate))
            stmt.setString(3, tags.serialize())
            stmt.setString(4, description)
            stmt.setInt(5, state.code)
            stmt.setLong(6, accountId)
            stmt.setLong(7, id)

            stmt.executeUpdate()
        }
    }

    fun getDocumentByExtId(accountId: Long, extId: String): Document {
        return connection.prepareStatement(
            """
            |SELECT 
            |    d.id as d_id,
            |    d.account_id d_account_id, 
            |    d.ext_id d_ext_id, 
            |    d.title d_title, 
            |    d.description d_description, 
            |    d.tags d_tags, 
            |    d.created d_created, 
            |    d.reference_date d_reference_date,
            |    d.state as d_state,
            |    f.id f_id,
            |    f.ext_id f_ext_id,
            |    f.account_id f_account_id,
            |    f.document_id f_document_id,
            |    f.filename f_filename,
            |    f.content_type f_content_type,
            |    f.content_extracted f_content_extracted,
            |    f.file_size f_file_size,
            |    f.created f_created,
            |    f.state f_state
            |FROM 
            |    document d LEFT OUTER JOIN file f ON (d.id = f.document_id) 
            |WHERE 
            |    d.ext_id=? AND 
            |    d.account_id=?
            """.trimMargin()
        )
            .use { stmt ->


                stmt.setString(1, extId)
                stmt.setLong(2, accountId)
                val res = stmt.executeQuery()
                var document: Document? = null
                val files = mutableListOf<FileRef>()

                while (res.next()) {
                    if (document == null) {
                        document = Document(
                            id = res.getLong("d_id"),
                            extId = res.getString("d_ext_id"),
                            title = res.getString("d_title"),
                            description = res.getString("d_description"),
                            tags = Tag.parse(res.getString("d_tags")),
                            created = LocalDateTime.parse(res.getString("d_created"), dtf),
                            referenceDate = LocalDateTime.parse(res.getString("d_reference_date"), dtf),
                            state = State.fromCode(res.getInt("d_state")),
                            files = emptyList()
                        )
                    }

                    val fid: Long = res.getLong("f_id")
                    val wasNull = res.wasNull()

                    if (!wasNull) {
                        files.add(
                            FileRef(
                                id = fid,
                                extId = res.getString("f_ext_id"),
                                accountId = res.getLong("f_account_id"),
                                documentId = res.getLong("f_document_id"),
                                filename = res.getString("f_filename"),
                                contentType = res.getString("f_content_type"),
                                contentExtracted = res.getString("f_content_extracted"),
                                fileSize = res.getLong("f_file_size"),
                                created = LocalDateTime.parse(res.getString("f_created"), dtf),
                                state = State.fromCode(res.getInt("f_state"))
                            )
                        )
                    }

                }

                requireNotNull(document)
                document.copy(files = files)
            }
    }

    fun loadFile(accountId: Long, extId: String): FileContent {
        return connection.prepareStatement(
            """
            |SELECT
            |   id,
            |   ext_id,
            |   filename,
            |   content_type,
            |   content_extracted,
            |   file_size,
            |   content
            |FROM
            |   file
            |WHERE
            |   account_id=? AND
            |   ext_id=?
        """.trimMargin()
        ).use { stmt ->
            stmt.setLong(1, accountId)
            stmt.setString(2, extId)
            val res = stmt.executeQuery()

            FileContent(
                id = res.getLong(1),
                extId = res.getString(2),
                filename = res.getString(3),
                contentType = res.getString(4),
                contentExtracted = res.getString(5),
                fileSize = res.getLong(6),
                content = res.getBytes(7)
            )
        }
    }
}